;**************************************************************************************
;*
;* saints.asm
;*
;**************************************************************************************
;*
;* multi-channel music, using pulse width modulation
;*
;**************************************************************************************

;*
;* device and config options
;*

		LIST		P=16F84

#include <p16f84.inc>

		__config	h'3FFA'

;*
;* definitions
;*

BLOCKSIZE	equ		d'4000'

;*
;* registers
;*

; note values/frequencies
note1		equ		h'0c'
note1hi		equ		h'0d'
note2		equ		h'0e'
note2hi		equ		h'0f'
note3		equ		h'10'
note3hi		equ		h'11'

noteptr		equ		h'12'		; song position
channel		equ		h'13'		; current channel
total		equ		h'14'		; total accumulated amplitude

val1		equ		h'15'
val1hi		equ		h'16'
val2		equ		h'17'
val2hi		equ		h'18'
val3		equ		h'19'
val3hi		equ		h'1a'

blockct		equ		h'1b'
blockcthi	equ		h'1c'

waitct		equ		h'1d'
waitct2		equ		h'1e'

;*
;* code
;*

		org		0

;**************************************************************************************
;*
;* start/intvec
;*
;**************************************************************************************
;*
;* CPU startup/interrupt entry pts
;*
;**************************************************************************************

start		clrf		INTCON
		goto		main
		nop
		nop

intvec		goto		intvec
		nop
		nop
		nop


;**************************************************************************************
;*
;* main
;*
;**************************************************************************************
;*
;*
;*
;**************************************************************************************

main
		;*
		;* Initialise IO ports
		;*

		bsf		STATUS,RP0		; select bank 1

		movlw		b'11111111'
		movwf		TRISA			; port A = all inputs
		clrf		TRISB			; port B = all outputs

		;*
		;* Initialise TMR0
		;*

		movlw		b'11001000'		; no prescale on tmr0
		movwf		OPTION_REG		; write into option register
		bcf		STATUS,RP0		; select bank 0

		;*
		;* init tune
		;*

start_tune
		clrf		noteptr			; reset song position

		;*
		;* main tune loop
		;*

tune_loop
		movlw		note1			; set ptr to note1
		movwf		FSR
		movlw		d'3'			; set channel count
		movwf		channel

		;*
		;* get frequencies for this block
		;*

channel_loop
		call		tunedata		; fetch note
		movwf		INDF			; store value
		btfsc		INDF,7			; test for bit 7 - end of tune marker
		goto		start_tune		; restart
		rlf		INDF,1			; get note value*2
		call		freqdata		; get high byte of frequency
		incf		FSR,1			; advance to second location
		movwf		INDF			; write high byte
		decf		FSR,1			; return to first location
		incf		INDF,1			; get note value*2+1
 		call		freqdata		; get low byte of frequency
		movwf	 	INDF			; write low byte
		incf		FSR,1			; advance ptr
		incf		FSR,1
		incf		noteptr,1		; advance note ptr
		decfsz		channel,1		; process all 3 channels
		goto		channel_loop

		;*
 		;* output BLOCKSIZE x 5 samples @ 44100
		;*

		movlw		(BLOCKSIZE >> 8)
		movwf		blockcthi
		movlw		(BLOCKSIZE & h'FF')
		movwf		blockct

block_loop	bcf		PORTB,0			; lower output
		clrf		total			; clear total

		movf		note1,0			; get lo inc
		addwf		val1,1			; add to lo
		btfsc		STATUS,C		; carry?
		incf		val1hi,1		; inc hi if so
		movf		note1hi,0		; get hi inc
		addwf		val1hi,1		; add to hi
		btfsc		val1hi,7		; test 32768 bit
		incf		total,1

		movf		note2,0			; get lo inc
		addwf		val2,1			; add to lo
		btfsc		STATUS,C		; carry?
		incf		val2hi,1		; inc hi if so
		movf		note2hi,0		; get hi inc
		addwf		val2hi,1		; add to hi
		btfsc		val2hi,7		; test 32768 bit
		incf		total,1

		movf		note3,0			; get lo inc
		addwf		val3,1			; add to lo
		btfsc		STATUS,C		; carry?
		incf		val3hi,1		; inc hi if so
		movf		note3hi,0		; get hi inc
		addwf		val3hi,1		; add to hi
		btfsc		val3hi,7		; test 32768 bit
		incf		total,1

		movlw		d'4'			; set wait ctr
		movwf		waitct
		decf		total,1			; total is 255 to 2

waitloop	bsf		PORTB,1
		btfsc		total,7			; test overflow
		bcf		PORTB,1

		; 23 ticks...
		movlw		d'7'
		movwf		waitct2
waitloop2	decfsz		waitct2,1
		goto		waitloop2

		decf		total,1
		decfsz		waitct,1
		goto		waitloop

		decfsz		blockct,1
		goto		block_loop
		decfsz		blockcthi,1
		goto		block_loop

		goto		tune_loop


;**************************************************************************************
;*
;* tunedata
;*
;**************************************************************************************
;*
;* table containing three channel note info for tune - fetch from noteptr
;*
;**************************************************************************************

		org		d'512'
tunedata	movlw		(tunedatatable >> 8)
		movwf		PCLATH
		movf		noteptr,0
		addlw		(tunedatatable & h'ff')
		movwf		PCL

tunedatatable
		#include	"tune.asm"


;**************************************************************************************
;*
;* freqdata
;*
;**************************************************************************************
;*
;* table for converting notes to frequencies - read from INDF
;*
;**************************************************************************************

		org		d'768'

freqdata	movlw		(freqdatatable >> 8)
		movwf		PCLATH
		movf		INDF,0
		addlw		(freqdatatable & h'ff')
		movwf		PCL

freqdatatable
		#include	"freq.asm"

;**************************************************************************************
;*
;*
;*
;**************************************************************************************
;*
;*
;*
;**************************************************************************************

;**************************************************************************************
;*
;*
;*
;**************************************************************************************
;*
;*
;*
;**************************************************************************************


		end

;*
;* data
;*

;*
;* end
;*

